@
@ GCat's Master System Emulator
@ Copyright (c) 2012 James Perry
@
@ This program is free software: you can redistribute it and/or modify
@ it under the terms of the GNU General Public License as published by
@ the Free Software Foundation, either version 3 of the License, or
@ (at your option) any later version.
@
@ This program is distributed in the hope that it will be useful,
@ but WITHOUT ANY WARRANTY; without even the implied warranty of
@ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
@ GNU General Public License for more details.
@
@ You should have received a copy of the GNU General Public License
@ along with this program.  If not, see <http://www.gnu.org/licenses/>.
@

@
@ Render a scanline
@
@ This is inlined into the end of line code in z80.S. On entry:
@  r0 - points to z80State
@  r3 - scanline number - this is only called if it's in range 64-256
@  r12 - points to screenState
@
@ r0 and r12 must be preserved on exit. All other registers can be
@ clobbered.
@
	sub	r3, r3, #64		@ get actual scanline number

	@
	@ For first loop:
	@  r1 - loop counter
	@  r2 - pointer into name table in video RAM
	@  r3 - scratch
	@  r4 - tile info word(s) from video RAM
	@  r5 - pointer to output line buffer
	@  r6 - output byte index
	@  r7 - holds rwt and urwt
	@  r8 - increment for r6
	@  r9 - paletteConverted pointer
	@  r10 - tileBitmaps pointer
	@  r11 - scratch
	@
	@ We can reuse r2, r5, r6, r7 for second tile loop as well
	@ Sprite code can also use r5
	@

	ldr	r5, [r12, #48]		@ screenBitmap
	add	r5, r5, r3, lsl #10	@ 1024 bytes per line

	ldrb	r6, [r12, #20]		@ output byte index is horizontal
					@ scroll (from register 8)

	ldrb	r2, [r12, #21]		@ vertical scrolling
	add	r2, r2, r3		@ work out scrolled row
	cmp	r2, #224		@ max 224
	subge	r2, r2, #224

	ldrb	r11, [r12, #12]		@ register 0
	tst	r11, #0x40
	beq	render_no_top_panel	@ test for fixed 2 rows at top
	cmp	r3, #16
	bge	render_no_top_panel	@ test for in top 2 rows
	mov	r2, r3			@ don't scroll in that case
	mov	r6, #0			@ no horizontal scroll either
render_no_top_panel:

	@ work out offsets within tile
	and	r7, r2, #7
	rsb	r11, r7, #7
	orr	r7, r7, r11, lsl #16	@ upside-down one goes in top word
	
	@ r2 is scrolled row. work out name table pointer
	ldrb	r11, [r12, #14]
	and	r11, r11, #14
	and	r2, r2, #0xf8
	mov	r2, r2, lsl #3
	orr	r2, r2, r11, lsl #10
	ldr	r11, [r12, #4]		@ VRAM pointer
	add	r2, r2, r11		@ nametable pointer now in r2

	push	{r2, r6, r7}		@ keep for second loop
	
	mov	r1, #32			@ render 32 tiles
render_tile_loop1:

	ldrh	r4, [r2], #2		@ read next tile info

	ldr	r9, [r12, #40]		@ paletteConverted
	tst	r4, #0x800
	addne	r9, r9, #64

	ldr	r10, [r12, #44]		@ tileBitmaps
	bic	r11, r4, #0xfe00	@ tile number
	add	r10, r10, r11, lsl #6
	tst	r4, #0x400		@ flip vertical?
	movne	r11, r7, lsr #16
	andeq	r11, r7, #7
	add	r10, r10, r11, lsl #3	@ r10 now source address

	@ flip horizontal?
	mov	r8, #1
	tst	r4, #0x200
	movne	r8, #0xff
	addne	r6, r6, #7

	and	r6, r6, #0xff
	
	@ read first 4 bytes
	ldr	r11, [r10], #4

	and	r3, r11, #0xf		@ get first pixel
	ldr	r3, [r9, r3, lsl #2]	@ lookup in palette
	str	r3, [r5, r6, lsl #2]	@ store in output
	add	r6, r6, r8
	and	r6, r6, #0xff

	mov	r3, r11, lsr #8
	and	r3, r3, #0xf		@ get next pixel
	ldr	r3, [r9, r3, lsl #2]	@ lookup in palette
	str	r3, [r5, r6, lsl #2]	@ store in output
	add	r6, r6, r8
	and	r6, r6, #0xff
	
	mov	r3, r11, lsr #16
	and	r3, r3, #0xf		@ get next pixel
	ldr	r3, [r9, r3, lsl #2]	@ lookup in palette
	str	r3, [r5, r6, lsl #2]	@ store in output
	add	r6, r6, r8
	and	r6, r6, #0xff
	
	mov	r3, r11, lsr #24
	and	r3, r3, #0xf		@ get next pixel
	ldr	r3, [r9, r3, lsl #2]	@ lookup in palette
	str	r3, [r5, r6, lsl #2]	@ store in output
	add	r6, r6, r8
	and	r6, r6, #0xff
	
	@ read next 4 bytes
	ldr	r11, [r10]

	and	r3, r11, #0xf		@ get first pixel
	ldr	r3, [r9, r3, lsl #2]	@ lookup in palette
	str	r3, [r5, r6, lsl #2]	@ store in output
	add	r6, r6, r8
	and	r6, r6, #0xff

	mov	r3, r11, lsr #8
	and	r3, r3, #0xf		@ get next pixel
	ldr	r3, [r9, r3, lsl #2]	@ lookup in palette
	str	r3, [r5, r6, lsl #2]	@ store in output
	add	r6, r6, r8
	and	r6, r6, #0xff
	
	mov	r3, r11, lsr #16
	and	r3, r3, #0xf		@ get next pixel
	ldr	r3, [r9, r3, lsl #2]	@ lookup in palette
	str	r3, [r5, r6, lsl #2]	@ store in output
	add	r6, r6, r8
	and	r6, r6, #0xff
	
	mov	r3, r11, lsr #24
	and	r3, r3, #0xf		@ get next pixel
	ldr	r3, [r9, r3, lsl #2]	@ lookup in palette
	str	r3, [r5, r6, lsl #2]	@ store in output
	add	r6, r6, r8
	and	r6, r6, #0xff

	tst	r4, #0x200
	addne	r6, r6, #9

	@ 8 columns fixed on right
	@ if this is enabled, when we get to column 24, need to:
	@  - update r7 to new rwt and urwt
	@  - update r2 to new nametable pointer
	cmp	r1, #9			@ not yet decremented
	bne	render_tile1_no8col
	ldrb	r3, [r12, #12]		@ reg 0
	tst	r3, #0x80
	beq	render_tile1_no8col
	
	@ let's do this thing
	ldrb	r3, [r12, #0]		@ scanline
	sub	r3, r3, #64
	and	r7, r3, #7
	rsb	r11, r7, #7
	orr	r7, r7, r11, lsl #16	@ got new rwt/urwt now
	
	ldrb	r11, [r12, #14]
	and	r11, r11, #14
	and	r2, r3, #0xf8
	mov	r2, r2, lsl #3
	orr	r2, r2, r11, lsl #10
	ldr	r11, [r12, #4]
	add	r2, r2, r11		@ new nametable pointer
	add	r2, r2, #48
	
render_tile1_no8col:

	subs	r1, r1, #1		@ loop back over 32 tiles
	bne	render_tile_loop1

@
@ Now draw sprites!
@
@ For sprite loop:
@   r1 = loop counter (0-63 this time)
@   r2 = sprite table pointer
@   r3 = screen line
@   r4 = sprites found so far
@   r5 = output buffer address
@   r6 = d8-15 = sprite height in pixels
@	 d0 = 16x16 sprites enabled
@        d1 = 8x16 sprites enabled
@        d2 = sprite tile '9th bit'
@	 d3 = shift all sprites right
@   r7 = sprite Y
@   r8 = sprite X / offset to write
@   r9 = palette pointer
@   r10 = source pointer
@   r11 = scratch
@
@ Remember: preserve r0, r5, r12!
@
	ldr	r3, [r12, #0]
	sub	r3, r3, #64		@ get screen line back

	mov	r6, #0x800		@ sprite height in pixels
	ldrh	r11, [r12, #12]		@ VDP reg 0 and 1
	tst	r11, #0x300
	movne	r6, #0x1000		@ tall sprites

	tst	r11, #0x100		@ 16x16 sprites
	orrne	r6, r6, #1
	tst	r11, #0x200		@ 8x16 sprites
	orrne	r6, r6, #2
	tst	r11, #8			@ shift all sprites right
	orrne	r6, r6, #8

	ldrb	r11, [r12, #18]		@ VDP reg 6
	tst	r11, #4
	orrne	r6, r6, #4		@ 9th bit

	ldrb	r2, [r12, #17]		@ VDP reg 5 - sprite table pointer
	and	r2, r2, #0x7e
	ldr	r11, [r12, #4]		@ VRAM
	add	r2, r11, r2, lsl #7	@ sprite table pointer
	
	mov	r4, #0			@ sprites on this line so far
	mov	r1, #0
render_sprite_loop:
	ldrb	r7, [r2, r1]		@ read sprite Y
	cmp	r7, #208		@ end marker?
	beq	render_sprite_loop_exit

	cmp	r7, r3
	bgt	render_sprite_loop_end	@ it's fully below this line
	add	r11, r7, r6, lsr #8
	cmp	r11, r3
	ble	render_sprite_loop_end	@ it's fully above this line

	add	r2, r2, r1, lsl #1
	ldrh	r8, [r2, #128]		@ read sprite X (low), tile (hi)
	sub	r2, r2, r1, lsl #1

	mov	r9, r8, lsr #8		@ get sprite tile number

	tst	r6, #8
	addne	r8, r8, #8		@ shift right if necessary
	and	r8, r8, #0xff		@ get sprite x

	tst 	r6, #2
	andne	r9, r9, #0xfe		@ round tile down for 8x16 sprites
	tst	r6, #4
	orrne	r9, r9, #0x100		@ add 9th bit

	tst	r6, #1
	bne	render_sprite_16x16
	
	sub	r3, r3, r7		@ get which row of sprite
	ldr	r10, [r12, #44]		@ tileBitmaps
	add	r10, r10, r9, lsl #6	@ point to tile
	add	r10, r10, r3, lsl #3	@ point to row within tile
	
	ldr	r9, [r12, #40]
	add	r9, r9, #64		@ sprite palette

	@ for drawing we need:
	@  r10 pointing to tileBitmap array (got)
	@  r11 and r3 as scratch (got)
	@  r9 pointing to palette (got)
	@  r5 pointing to dest array (got)
	@  r8 as offset to write (got)
	@ read first 4 bytes
	ldr	r11, [r10], #4

	ands	r3, r11, #0xf		@ get first pixel
	ldrne	r3, [r9, r3, lsl #2]	@ lookup in palette
	strne	r3, [r5, r8, lsl #2]	@ store in output
	add	r8, r8, #1
	and	r8, r8, #0xff

	mov	r3, r11, lsr #8
	ands	r3, r3, #0xf		@ get next pixel
	ldrne	r3, [r9, r3, lsl #2]	@ lookup in palette
	strne	r3, [r5, r8, lsl #2]	@ store in output
	add	r8, r8, #1
	and	r8, r8, #0xff
	
	mov	r3, r11, lsr #16
	ands	r3, r3, #0xf		@ get next pixel
	ldrne	r3, [r9, r3, lsl #2]	@ lookup in palette
	strne	r3, [r5, r8, lsl #2]	@ store in output
	add	r8, r8, #1
	and	r8, r8, #0xff
	
	mov	r3, r11, lsr #24
	ands	r3, r3, #0xf		@ get next pixel
	ldrne	r3, [r9, r3, lsl #2]	@ lookup in palette
	strne	r3, [r5, r8, lsl #2]	@ store in output
	add	r8, r8, #1
	and	r8, r8, #0xff
	
	@ read next 4 bytes
	ldr	r11, [r10]
	ands	r3, r11, #0xf		@ get first pixel
	ldrne	r3, [r9, r3, lsl #2]	@ lookup in palette
	strne	r3, [r5, r8, lsl #2]	@ store in output
	add	r8, r8, #1
	and	r8, r8, #0xff

	mov	r3, r11, lsr #8
	ands	r3, r3, #0xf		@ get next pixel
	ldrne	r3, [r9, r3, lsl #2]	@ lookup in palette
	strne	r3, [r5, r8, lsl #2]	@ store in output
	add	r8, r8, #1
	and	r8, r8, #0xff
	
	mov	r3, r11, lsr #16
	ands	r3, r3, #0xf		@ get next pixel
	ldrne	r3, [r9, r3, lsl #2]	@ lookup in palette
	strne	r3, [r5, r8, lsl #2]	@ store in output
	add	r8, r8, #1
	and	r8, r8, #0xff
	
	mov	r3, r11, lsr #24
	ands	r3, r3, #0xf		@ get next pixel
	ldrne	r3, [r9, r3, lsl #2]	@ lookup in palette
	strne	r3, [r5, r8, lsl #2]	@ store in output
	add	r8, r8, #1
	and	r8, r8, #0xff

	b	render_sprite_done
	
render_sprite_16x16:
	sub	r3, r3, r7		@ get which row of sprite
	and	r3, r3, #14
	ldr	r10, [r12, #44]		@ tileBitmaps
	add	r10, r10, r9, lsl #6	@ point to tile
	add	r10, r10, r3, lsl #2	@ point to row within tile
	
	ldr	r9, [r12, #40]
	add	r9, r9, #64		@ sprite palette

	@ read first 4 bytes
	ldr	r11, [r10], #4

	ands	r3, r11, #0xf		@ get first pixel
	ldrne	r3, [r9, r3, lsl #2]	@ lookup in palette
	strne	r3, [r5, r8, lsl #2]	@ store in output
	add	r8, r8, #1
	and	r8, r8, #0xff
	strne	r3, [r5, r8, lsl #2]	@ store in output
	add	r8, r8, #1
	and	r8, r8, #0xff

	mov	r3, r11, lsr #8
	ands	r3, r3, #0xf		@ get next pixel
	ldrne	r3, [r9, r3, lsl #2]	@ lookup in palette
	strne	r3, [r5, r8, lsl #2]	@ store in output
	add	r8, r8, #1
	and	r8, r8, #0xff
	strne	r3, [r5, r8, lsl #2]	@ store in output
	add	r8, r8, #1
	and	r8, r8, #0xff
	
	mov	r3, r11, lsr #16
	ands	r3, r3, #0xf		@ get next pixel
	ldrne	r3, [r9, r3, lsl #2]	@ lookup in palette
	strne	r3, [r5, r8, lsl #2]	@ store in output
	add	r8, r8, #1
	and	r8, r8, #0xff
	strne	r3, [r5, r8, lsl #2]	@ store in output
	add	r8, r8, #1
	and	r8, r8, #0xff
	
	mov	r3, r11, lsr #24
	ands	r3, r3, #0xf		@ get next pixel
	ldrne	r3, [r9, r3, lsl #2]	@ lookup in palette
	strne	r3, [r5, r8, lsl #2]	@ store in output
	add	r8, r8, #1
	and	r8, r8, #0xff
	strne	r3, [r5, r8, lsl #2]	@ store in output
	add	r8, r8, #1
	and	r8, r8, #0xff
	
	@ read next 4 bytes
	ldr	r11, [r10]
	ands	r3, r11, #0xf		@ get first pixel
	ldrne	r3, [r9, r3, lsl #2]	@ lookup in palette
	strne	r3, [r5, r8, lsl #2]	@ store in output
	add	r8, r8, #1
	and	r8, r8, #0xff
	strne	r3, [r5, r8, lsl #2]	@ store in output
	add	r8, r8, #1
	and	r8, r8, #0xff

	mov	r3, r11, lsr #8
	ands	r3, r3, #0xf		@ get next pixel
	ldrne	r3, [r9, r3, lsl #2]	@ lookup in palette
	strne	r3, [r5, r8, lsl #2]	@ store in output
	add	r8, r8, #1
	and	r8, r8, #0xff
	strne	r3, [r5, r8, lsl #2]	@ store in output
	add	r8, r8, #1
	and	r8, r8, #0xff
	
	mov	r3, r11, lsr #16
	ands	r3, r3, #0xf		@ get next pixel
	ldrne	r3, [r9, r3, lsl #2]	@ lookup in palette
	strne	r3, [r5, r8, lsl #2]	@ store in output
	add	r8, r8, #1
	and	r8, r8, #0xff
	strne	r3, [r5, r8, lsl #2]	@ store in output
	add	r8, r8, #1
	and	r8, r8, #0xff
	
	mov	r3, r11, lsr #24
	ands	r3, r3, #0xf		@ get next pixel
	ldrne	r3, [r9, r3, lsl #2]	@ lookup in palette
	strne	r3, [r5, r8, lsl #2]	@ store in output
	add	r8, r8, #1
	and	r8, r8, #0xff
	strne	r3, [r5, r8, lsl #2]	@ store in output
	add	r8, r8, #1
	and	r8, r8, #0xff

render_sprite_done:	

	ldr	r3, [r12, #0]
	sub	r3, r3, #64		@ get screen line back
	add	r4, r4, #1		@ one more sprite on the line
render_sprite_loop_end:
	cmp	r4, #8
	bge	render_sprite_loop_exit
	add	r1, r1, #1
	cmp	r1, #64
	blt	render_sprite_loop
render_sprite_loop_exit:

	and	r4, r4, #8
	ldrb	r2, [r12, #32]		@ video status
	orr	r2, r2, r4, lsl #3	@ sprite overflow flag
	strb	r2, [r12, #32]

@
@ Now draw tiles that appear in front of sprites!
@
@ Still need to setup r1, r9, r10
@
	pop	{r2, r6, r7}		@ get back preserved registers

	mov	r1, #32			@ render 32 tiles
render_tile_loop2:

	ldrh	r4, [r2], #2		@ read next tile info

	tst	r4, #0x1000		@ in front of sprites?
	addeq	r6, r6, #8
	beq	render_skip_tile	@ skip if not
	
	ldr	r9, [r12, #40]		@ paletteConverted
	tst	r4, #0x800
	addne	r9, r9, #64

	ldr	r10, [r12, #44]		@ tileBitmaps
	bic	r11, r4, #0xfe00	@ tile number
	add	r10, r10, r11, lsl #6
	tst	r4, #0x400		@ flip vertical?
	movne	r11, r7, lsr #16
	andeq	r11, r7, #7
	add	r10, r10, r11, lsl #3	@ r10 now source address

	@ flip horizontal?
	mov	r8, #1
	tst	r4, #0x200
	movne	r8, #0xff
	addne	r6, r6, #7

	and	r6, r6, #0xff
	
	@ read first 4 bytes
	ldr	r11, [r10], #4

	ands	r3, r11, #0xf		@ get first pixel
	ldrne	r3, [r9, r3, lsl #2]	@ lookup in palette
	strne	r3, [r5, r6, lsl #2]	@ store in output
	add	r6, r6, r8
	and	r6, r6, #0xff

	mov	r3, r11, lsr #8
	ands	r3, r3, #0xf		@ get next pixel
	ldrne	r3, [r9, r3, lsl #2]	@ lookup in palette
	strne	r3, [r5, r6, lsl #2]	@ store in output
	add	r6, r6, r8
	and	r6, r6, #0xff
	
	mov	r3, r11, lsr #16
	ands	r3, r3, #0xf		@ get next pixel
	ldrne	r3, [r9, r3, lsl #2]	@ lookup in palette
	strne	r3, [r5, r6, lsl #2]	@ store in output
	add	r6, r6, r8
	and	r6, r6, #0xff
	
	mov	r3, r11, lsr #24
	ands	r3, r3, #0xf		@ get next pixel
	ldrne	r3, [r9, r3, lsl #2]	@ lookup in palette
	strne	r3, [r5, r6, lsl #2]	@ store in output
	add	r6, r6, r8
	and	r6, r6, #0xff
	
	@ read next 4 bytes
	ldr	r11, [r10]
	ands	r3, r11, #0xf		@ get first pixel
	ldrne	r3, [r9, r3, lsl #2]	@ lookup in palette
	strne	r3, [r5, r6, lsl #2]	@ store in output
	add	r6, r6, r8
	and	r6, r6, #0xff

	mov	r3, r11, lsr #8
	ands	r3, r3, #0xf		@ get next pixel
	ldrne	r3, [r9, r3, lsl #2]	@ lookup in palette
	strne	r3, [r5, r6, lsl #2]	@ store in output
	add	r6, r6, r8
	and	r6, r6, #0xff
	
	mov	r3, r11, lsr #16
	ands	r3, r3, #0xf		@ get next pixel
	ldrne	r3, [r9, r3, lsl #2]	@ lookup in palette
	strne	r3, [r5, r6, lsl #2]	@ store in output
	add	r6, r6, r8
	and	r6, r6, #0xff
	
	mov	r3, r11, lsr #24
	ands	r3, r3, #0xf		@ get next pixel
	ldrne	r3, [r9, r3, lsl #2]	@ lookup in palette
	strne	r3, [r5, r6, lsl #2]	@ store in output
	add	r6, r6, r8
	and	r6, r6, #0xff

	tst	r4, #0x200
	addne	r6, r6, #9

	@ 8 columns fixed on right
	@ if this is enabled, when we get to column 24, need to:
	@  - update r7 to new rwt and urwt
	@  - update r2 to new nametable pointer
	cmp	r1, #9			@ not yet decremented
	bne	render_tile2_no8col
	ldrb	r3, [r12, #12]		@ reg 0
	tst	r3, #0x80
	beq	render_tile2_no8col
	
	@ let's do this thing
	ldrb	r3, [r12, #0]		@ scanline
	sub	r3, r3, #64
	and	r7, r3, #7
	rsb	r11, r7, #7
	orr	r7, r7, r11, lsl #16	@ got new rwt/urwt now
	
	ldrb	r11, [r12, #14]
	and	r11, r11, #14
	and	r2, r3, #0xf8
	mov	r2, r2, lsl #3
	orr	r2, r2, r11, lsl #10
	ldr	r11, [r12, #4]
	add	r2, r2, r11		@ new nametable pointer
	add	r2, r2, #48
	
render_tile2_no8col:

render_skip_tile:
	subs	r1, r1, #1		@ loop back over 32 tiles
	bne	render_tile_loop2

